# nervous_system.py
import threading
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from sensory import Sensory
from brain import Brain


class NervousSystem:
    """Wraps the Brain, handles token growth, generation, and training."""

    def __init__(self, device: str = "cuda"):
        self.device = torch.device(
            device if torch.cuda.is_available() else "cpu"
        )
        self.sensory = Sensory()
        vocab_size = len(self.sensory.stoi)
        self.brain = Brain(vocab_size).to(self.device)

        # disable any inplace ops in the model
        for m in self.brain.modules():
            if hasattr(m, "inplace"):
                m.inplace = False  # ensure no in-place ReLUs, etc.

        self.optimizer = optim.Adam(self.brain.parameters(), lr=1e-4)
        self.criterion = nn.CrossEntropyLoss(ignore_index=0)
        self.meta_steps = 0

        # ← NEW: lock to serialize all training calls
        self._train_lock = threading.Lock()

    def _resize_embeddings(self) -> None:
        """
        Resize token & output embeddings entirely on-device
        to match the current vocab size, avoiding any CPU↔GPU copies.
        """
        device = self.device
        new_size = len(self.sensory.stoi)

        # ─── Resize token embeddings ────────────────────────────────────────
        old_emb = self.brain.token_emb
        old_num, dim = old_emb.num_embeddings, old_emb.embedding_dim

        # allocate new on same device
        new_emb = nn.Embedding(new_size, dim).to(device)
        # copy existing weights
        with torch.no_grad():
            new_emb.weight[:old_num].copy_(old_emb.weight)
        self.brain.token_emb = new_emb

        # ─── Resize output head ─────────────────────────────────────────────
        old_out = self.brain.fc_out
        out_dim, old_out_feat = old_out.in_features, old_out.out_features

        new_out = nn.Linear(out_dim, new_size).to(device)
        with torch.no_grad():
            new_out.weight[:old_out_feat].copy_(old_out.weight)
            new_out.bias[:old_out_feat].copy_(old_out.bias)
        self.brain.fc_out = new_out

    def generate(
        self,
        prompt: str,
        max_len: int = 50,
        temperature: float = 0.8,
        top_p: float = 0.9,
    ) -> str:
        """Autoregressive nucleus sampling with proper cloning to avoid aliasing."""
        self.brain.eval()
        eos_id = self.sensory.stoi.get("<eos>")

        raw_ids = self.sensory.encode(prompt, grow=False)
        max_ctx = self.brain.max_seq_len - 1
        if len(raw_ids) > max_ctx:
            raw_ids = raw_ids[-max_ctx:]

        input_ids = torch.tensor(
            raw_ids, dtype=torch.long, device=self.device
        ).unsqueeze(0)

        generated = []
        for _ in range(max_len):
            logits = self.brain(input_ids)[:, -1, :] / temperature
            probs = F.softmax(logits, dim=-1)

            # top-p (nucleus) filtering
            sorted_p, sorted_idx = torch.sort(probs, descending=True)
            cum_p = torch.cumsum(sorted_p, dim=-1)
            mask = cum_p > top_p
            # clone the slice before writing to avoid overlap
            mask_shift = mask[..., :-1].clone()
            mask[..., 1:] = mask_shift
            sorted_p[mask] = 0
            sorted_p = sorted_p / sorted_p.sum(dim=-1, keepdim=True)

            next_tok = sorted_idx[0, torch.multinomial(sorted_p[0], 1)]
            t_id = next_tok.item()

            if eos_id is not None and t_id == eos_id:
                break

            generated.append(t_id)
            input_ids = torch.cat([input_ids, next_tok.unsqueeze(0)], dim=1)

            if input_ids.size(1) > self.brain.max_seq_len:
                input_ids = input_ids[:, -self.brain.max_seq_len:]

        return self.sensory.decode(generated)

    def train(self, user_text: str, bot_text: str) -> None:
        """
        On-the-fly self-supervised training from a user↔bot exchange.
        Serialized by a threading.Lock to avoid in-place grad conflicts.
        """
        with self._train_lock:  # ← NEW: no two threads can enter here at once
            # 1) grow vocab & resize embeddings
            for txt in (user_text, bot_text):
                _ = self.sensory.encode(txt, grow=True)
            self._resize_embeddings()

            # 2) ensure a <sep> token for concatenation
            if "<sep>" not in self.sensory.stoi:
                idx = len(self.sensory.stoi)
                self.sensory.stoi["<sep>"] = idx
                self.sensory.itos[idx] = "<sep>"
                self._resize_embeddings()

            # 3) build input/output IDs
            combined = f"{user_text} <sep> {bot_text}"
            ids = torch.tensor(
                self.sensory.encode(combined, grow=False),
                dtype=torch.long,
                device=self.device,
            ).unsqueeze(0)

            if ids.size(1) < 2:
                return

            inputs = ids[:, :-1]
            targets = ids[:, 1:]

            # 4) forward + backward + step
            self.brain.train()
            logits = self.brain(inputs)
            loss = self.criterion(
                logits.view(-1, logits.size(-1)), targets.view(-1)
            )
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            # 5) optional meta learning step adjustment
            self.meta_steps += 1
            if self.meta_steps % 100 == 0:
                for g in self.optimizer.param_groups:
                    g["lr"] *= 1.1
